#!/usr/bin/env python

import sys, string


class NotImplemented(Exception):
    pass


class GenericMatrix:

    def __init__(self, query, target):

        self.query_seq  = query
        self.target_seq = target

        self.query_len  = len(self.query_seq)
        self.target_len = len(self.target_seq)

        self.score_matrix = []
        self.dir_matrix   = []

        self.query_seq_align  = []
        self.target_seq_align = []
        self.match_seq_align  = []

        self.max_row   = 0
        self.max_col   = 0
        self.max_score = None

        for i in range(self.getTargetLength() ):
            self.score_matrix.append([None]*self.getQueryLength() )
            self.dir_matrix.append([None]*self.getQueryLength() )


    def traceBack(self, row, col, sw = False):

        while row >= 0 or col >= 0:

            score, direction = self.getCell(row, col)

            if sw is True and score == 0:
                break

            if not direction:

                query_res  = self.getQuerySequence()[col]
                target_res = self.getTargetSequence()[row]
                match_res  = None

                if query_res == target_res:
                    match_res = query_res

                elif self.matchScore(row, col) > 0:
                    match_res = '+'

                else:
                    match_res = ' '

                self.query_seq_align.append(query_res)
                self.target_seq_align.append(target_res)
                self.match_seq_align.append(match_res)

                row = row - 1
                col = col - 1

            elif direction < 0:

                for i in range(-direction):
                    self.query_seq_align.append('-')
                    self.target_seq_align.append(self.getTargetSequence()[row - i])
                    self.match_seq_align.append(' ')

                row = row + direction

            elif direction > 0:

                for i in range(direction):
                    self.query_seq_align.append(self.getQuerySequence()[col - i])
                    self.target_seq_align.append('-')
                    self.match_seq_align.append(' ')

                col = col - direction


    def printAlign(self):

        while True:

            try:
                res = self.query_seq_align.pop()
            except IndexError:
                break

            sys.stdout.write(res)

        print

        while True:

            try:
                res = self.match_seq_align.pop()
            except IndexError:
                break

            sys.stdout.write(res)

        print

        while True:

            try:
                res = self.target_seq_align.pop()
            except IndexError:
                break

            sys.stdout.write(res)

        print


    def printMatrices(self):

        for row in self.score_matrix:

            for col in row:
                print '%4d' % (int(col),),

            print

        print

        for row in self.dir_matrix:

            for col in row:
                print '%4d' % (int(col),),

            print

        print


    def getQuerySequence(self):

        return self.query_seq


    def getTargetSequence(self):

        return self.target_seq


    def getQueryLength(self):

        return self.query_len


    def getTargetLength(self):

        return self.target_len


    def getCell(self, row, col):

        return self.score_matrix[row][col], self.dir_matrix[row][col]


    def getScore(self, row, col):

        return self.getCell(row, col)[0]


    def getDirection(self, row, col):

        return self.getCell(row, col)[1]


    def setCell(self, row, col, score, direction):

        if self.max_score is None or score > self.max_score:
            self.max_score = score
            self.max_row   = row
            self.max_col   = col

        self.score_matrix[row][col] = score
        self.dir_matrix[row][col]   = direction


class SmithWatermanMatrix(GenericMatrix):

    def __init__(self, query, target):

        GenericMatrix.__init__(self, query, target)


    def setCell(self, row, col, score, direction):

        if score < 0.0:
            score = 0.0

        GenericMatrix.setCell(self, row, col, score, direction)


class MonotonicallyIncreasingGapAlgorithm:

    def __init__(self):
        pass


    def run(self):

        self.setCell(0, 0, self.matchScore(0, 0), 0)

        score_0, dir_0 = self.getCell(0, 0)

        for row in range(1, self.target_len):
            self.setCell(row, 0, score_0 - self.verticalGapScore(0, 0, row), -row)

        horiz_gap_start = [0]*self.getTargetLength()

        for col in range(1, self.query_len):
            self.setCell(0, col, score_0 - self.horizontalGapScore(0, 0, col), col)

        for col in range(1, self.query_len):

            vert_gap_start = 0

            for row in range(1, self.target_len):

                score = self.getScore(row - 1, col - 1) + self.matchScore(row, col)

                direction = 0

                horiz_gap_score = \
                    self.getScore(row, horiz_gap_start[row]) - \
                    self.horizontalGapScore(row, horiz_gap_start[row], col - horiz_gap_start[row])

                if horiz_gap_score > score:
                    score = horiz_gap_score
                    direction = col - horiz_gap_start[row]

                vert_gap_score = \
                    self.getScore(vert_gap_start, col) - \
                    self.verticalGapScore(vert_gap_start, col, row - vert_gap_start)

                if vert_gap_score > score:
                    score = vert_gap_score
                    direction = -(row - vert_gap_start)

                self.setCell(row, col, score, direction)

                score, direction = self.getCell(row, col)

                new_horiz_gap_score = score - self.horizontalGapScore(row, col, 1)

                old_horiz_gap_score = \
                    self.getScore(row, horiz_gap_start[row]) - \
                    self.horizontalGapScore(row, horiz_gap_start[row], col - horiz_gap_start[row] + 1)

                if new_horiz_gap_score > old_horiz_gap_score:
                    horiz_gap_start[row] = col

                new_vert_gap_score = score - self.verticalGapScore(row, col, 1)

                old_vert_gap_score = \
                    self.getScore(vert_gap_start, col) - \
                    self.verticalGapScore(vert_gap_start, col, row - vert_gap_start + 1)

                if new_vert_gap_score > old_vert_gap_score:
                    vert_gap_start = row


class AffineGapScore:

    def __init__(self, gap_open, gap_extend):

        self.gap_open   = gap_open
        self.gap_extend = gap_extend


    def getGapOpen(self):
        return self.gap_open


    def getGapExtend(self):
        return self.gap_extend

    
    def horizontalGapScore(self, row, col, length):
        return self.getGapOpen() + self.getGapExtend()*length


    verticalGapScore = horizontalGapScore


class MatrixScore:

    def __init__(self, matrix):

        self.matrix = matrix


    def matchScore(self, row, col):

        query_res  = self.getQuerySequence()[col]
        target_res = self.getTargetSequence()[row]

        return self.matrix.getScore(query_res, target_res)


class ProfileScore:

    def __init__(self, profile):

        self.profile = profile


    def matchScore(self, row, col):

        target_res = self.getTargetSequence()[row]

        return self.matrix.getScore(col, target_res)


class ScoreMatrix:

    def __init__(self, matrix_data, indirect = None, default_col = None):

        self.matrix_data = matrix_data

        if indirect:
            self.getScore = self._indirectGetScore

            self.indirect = {}

            for col in indirect:
                if default_col == col:
                    self.default = len(self.indirect)

                self.indirect[col] = len(self.indirect)

        else:
            self.getScore = self._rawGetScore


    def _indirectGetScore(self, query_res, target_res):

        try:
            query_res_in = self.indirect[query_res]

        except:
            query_res_in = self.default

        try:
            target_res_in = self.indirect[target_res]

        except:
            target_res_in = self.default

        return self.matrix_data[query_res_in][target_res_in]


    def _rawGetScore(self, query_res, target_res):

        return self.matrix_data[query_res][target_res]


class NeedleManWunsch(GenericMatrix,
                      AffineGapScore,
                      MonotonicallyIncreasingGapAlgorithm):

    def __init__(self, query, target, gap_open, gap_extend):

        GenericMatrix.__init__(self, query, target)
        AffineGapScore.__init__(self, gap_open, gap_extend)
        MonotonicallyIncreasingGapAlgorithm.__init__(self)


class SmithWaterman(SmithWatermanMatrix,
                    AffineGapScore,
                    MonotonicallyIncreasingGapAlgorithm):

    def __init__(self, query, target, gap_open, gap_extend):

        SmithWatermanMatrix.__init__(self, query, target)
        AffineGapScore.__init__(self, gap_open, gap_extend)
        MonotonicallyIncreasingGapAlgorithm.__init__(self)


class MatrixNeedleManWunsch(NeedleManWunsch,
                            MatrixScore):

    def __init__(self, query, target, gap_open, gap_extend, matrix):

        NeedleManWunsch.__init__(self, query, target, gap_open, gap_extend)
        MatrixScore.__init__(self, matrix)


class MatrixSmithWaterman(SmithWaterman,
                          MatrixScore):

    def __init__(self, query, target, gap_open, gap_extend, matrix):

        SmithWaterman.__init__(self, query, target, gap_open, gap_extend)
        MatrixScore.__init__(self, matrix)


class ProfileNeedleManWunsch(NeedleManWunsch,
                             ProfileScore):

    def __init__(self, query, target, gap_open, gap_extend, profile):

        NeedleManWunsch.__init__(self, query, target, gap_open, gap_extend)
        ProfileScore.__init__(self, profile)


class ProfileSmithWaterman(SmithWaterman,
                           ProfileScore):

    def __init__(self, query, target, gap_open, gap_extend, profile):

        SmithWaterman.__init__(self, query, target, gap_open, gap_extend)
        ProfileScore.__init__(self, profile)


def readSeq(file):

    f = open(file, 'r')

    line = f.readline()

    titleLine = None

    name = None
    desc = None

    if line:
        titleLine = line

        titleCols = titleLine.split()

        name = titleCols[0][:1]

        desc = titleCols[1:]

        #print '[', name, '] [', desc, "]\n"

    seq = ''

    while line:

        line = f.readline().strip()
        seq = seq + line

    f.close()

    return(name, desc, seq)


def readMatrix(file):

    f = open(file, 'r')

    line = f.readline()

    while line[0] == '#':
        line = f.readline()

    residues = []

    residues = line.split()
    residue_str = str.join('', residues)

    all_scores = []

    i = 0

    while line:
        line = f.readline()
        scores = line.split()

        if len(scores) == 0:
            break

        res = scores.pop(0)

        if residue_str[i] != res:
            print "Incorrect residue"

        #scores2 = [int(j) for j in scores]

        all_scores.append([string.atoi(j) for j in scores])

        i = i + 1

    f.close()

    return(all_scores, residue_str)


if __name__ == '__main__':

    #blosum62 = ScoreMatrix([
    #    [  4, -1, -2, -2,  0, -1, -1,  0, -2, -1, -1, -1, -1, -2, -1,  1,  0, -3, -2,  0, -2, -1,  0, -4 ],
    #    [ -1,  5,  0, -2, -3,  1,  0, -2,  0, -3, -2,  2, -1, -3, -2, -1, -1, -3, -2, -3, -1,  0, -1, -4 ],
    #    [ -2,  0,  6,  1, -3,  0,  0,  0,  1, -3, -3,  0, -2, -3, -2,  1,  0, -4, -2, -3,  3,  0, -1, -4 ],
    #    [ -2, -2,  1,  6, -3,  0,  2, -1, -1, -3, -4, -1, -3, -3, -1,  0, -1, -4, -3, -3,  4,  1, -1, -4 ],
    #    [  0, -3, -3, -3,  9, -3, -4, -3, -3, -1, -1, -3, -1, -2, -3, -1, -1, -2, -2, -1, -3, -3, -2, -4 ],
    #    [ -1,  1,  0,  0, -3,  5,  2, -2,  0, -3, -2,  1,  0, -3, -1,  0, -1, -2, -1, -2,  0,  3, -1, -4 ],
    #    [ -1,  0,  0,  2, -4,  2,  5, -2,  0, -3, -3,  1, -2, -3, -1,  0, -1, -3, -2, -2,  1,  4, -1, -4 ],
    #    [  0, -2,  0, -1, -3, -2, -2,  6, -2, -4, -4, -2, -3, -3, -2,  0, -2, -2, -3, -3, -1, -2, -1, -4 ],
    #    [ -2,  0,  1, -1, -3,  0,  0, -2,  8, -3, -3, -1, -2, -1, -2, -1, -2, -2,  2, -3,  0,  0, -1, -4 ],
    #    [ -1, -3, -3, -3, -1, -3, -3, -4, -3,  4,  2, -3,  1,  0, -3, -2, -1, -3, -1,  3, -3, -3, -1, -4 ],
    #    [ -1, -2, -3, -4, -1, -2, -3, -4, -3,  2,  4, -2,  2,  0, -3, -2, -1, -2, -1,  1, -4, -3, -1, -4 ],
    #    [ -1,  2,  0, -1, -3,  1,  1, -2, -1, -3, -2,  5, -1, -3, -1,  0, -1, -3, -2, -2,  0,  1, -1, -4 ],
    #    [ -1, -1, -2, -3, -1,  0, -2, -3, -2,  1,  2, -1,  5,  0, -2, -1, -1, -1, -1,  1, -3, -1, -1, -4 ],
    #    [ -2, -3, -3, -3, -2, -3, -3, -3, -1,  0,  0, -3,  0,  6, -4, -2, -2,  1,  3, -1, -3, -3, -1, -4 ],
    #    [ -1, -2, -2, -1, -3, -1, -1, -2, -2, -3, -3, -1, -2, -4,  7, -1, -1, -4, -3, -2, -2, -1, -2, -4 ],
    #    [  1, -1,  1,  0, -1,  0,  0,  0, -1, -2, -2,  0, -1, -2, -1,  4,  1, -3, -2, -2,  0,  0,  0, -4 ],
    #    [  0, -1,  0, -1, -1, -1, -1, -2, -2, -1, -1, -1, -1, -2, -1,  1,  5, -2, -2,  0, -1, -1,  0, -4 ],
    #    [ -3, -3, -4, -4, -2, -2, -3, -2, -2, -3, -2, -3, -1,  1, -4, -3, -2, 11,  2, -3, -4, -3, -2, -4 ],
    #    [ -2, -2, -2, -3, -2, -1, -2, -3,  2, -1, -1, -2, -1,  3, -3, -2, -2,  2,  7, -1, -3, -2, -1, -4 ],
    #    [  0, -3, -3, -3, -1, -2, -2, -3, -3,  3,  1, -2,  1, -1, -2, -2,  0, -3, -1,  4, -3, -2, -1, -4 ],
    #    [ -2, -1,  3,  4, -3,  0,  1, -1,  0, -3, -4,  0, -3, -3, -2,  0, -1, -4, -3, -3,  4,  1, -1, -4 ],
    #    [ -1,  0,  0,  1, -3,  3,  4, -2,  0, -3, -3,  1, -1, -3, -1,  0, -1, -3, -2, -2,  1,  4, -1, -4 ],
    #    [  0, -1, -1, -1, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2,  0,  0, -2, -1, -1, -1, -1, -1, -4 ],
    #    [ -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4,  1 ]
    #    ], 'ARNDCQEGHILKMFPSTWYVBZX*', '*')

    matrix = readMatrix('/nfs/team79/cjp/data/BLOSUM62.txt')

    blosum62 = ScoreMatrix(matrix[0], matrix[1], '*')

    new_seq1 = readSeq('/nfs/team79/cjp/data/new_test1.faa')
    new_seq2 = readSeq('/nfs/team79/cjp/data/new_test2.faa')
    new_seq3 = readSeq('/nfs/team79/cjp/data/new_test3.faa')

    nw_mat = MatrixNeedleManWunsch(new_seq1[2], new_seq3[2], 11, 1, blosum62)

    nw_mat.run()

    nw_mat.traceBack(nw_mat.target_len - 1, nw_mat.query_len - 1)

    print
    nw_mat.printAlign()
    print

    sw_mat = MatrixSmithWaterman(new_seq1[2], new_seq3[2], 11, 1, blosum62)

    sw_mat.run()

    sw_mat.traceBack(sw_mat.max_row, sw_mat.max_col, True)

    print
    sw_mat.printAlign()
    print

    #sw_mat.printMatrices()

    print sw_mat.max_score, 'at', sw_mat.max_row, ',', sw_mat.max_col
